# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import unohelper
import locale
import traceback
from itertools import groupby
from operator import itemgetter
from collections import defaultdict
from com.sun.star.xml import AttributeData
from com.sun.star.awt import XDialogEventHandler
from com.sun.star.container import NoSuchElementException
from com.sun.star.text.ControlCharacter import PARAGRAPH_BREAK
from com.sun.star.beans import PropertyValue, UnknownPropertyException
from com.sun.star.script import CannotConvertException
from com.sun.star.container import NoSuchElementException
from com.sun.star.lang import IllegalArgumentException

# some tweaks are needed
import sys
locale.setlocale(locale.LC_ALL, '')
if sys.version_info < (3, ):
    reload(sys)
    sys.setdefaultencoding(sys.getfilesystemencoding())

class Entry():
    def __init__(self):
        self.pages = set()
        self.ismain = False

    def add(self, page):
        self.pages.add(page)


class IndexCreator(unohelper.Base, XDialogEventHandler):
    SEP = '|'   # separator for key entries
    ISMAIN = '_'
    MAXKEYS = 2
    PATTERN = '{} \t '
    CONTENT_SUFFIX = "_index_content"
    HEADER_SUFFIX = "_index_header"
    DEFAULTSTYLES = {'title': "User Index Heading", 0: "Index Separator", 1: "User Index 1", 2: "User Index 2", 3: "User Index 3"}    # adapt on MAXKEYS changed
    LOCALES = dict(TITLE = {'it': "Indici analitici", 'fr': 'Index lexicaux', 'en': 'Alphabetical indexes'},
                   LBL_CHOICE = {'it': "Schegli l'indice ", 'fr': "Choisir l'index ", 'en': "Choose index "},
                   LBL_TITLE = {'it': "Titolo ", 'fr': "Titre ", 'en': "Title "},
                   LBL_OPTIONS = {'it': "Opzioni", 'fr': "Options", 'en': "Options"},
                   LBL_INSERT = {'it': "INSERISCI", 'fr': 'INSÉRER', 'en': "INSERT"},
                   LBL_UPDATE = {'it': "AGGIORNA", 'fr': 'METTRE À JOUR', 'en': "UPDATE"},
                   LBL_DELETE = {'it': "CANCELLA", 'fr': 'SUPPRIMER', 'en': "DELETE"},
                   LBL_CANCEL = {'it': "A~NNULA", 'fr': 'A~NNULER', 'en': "CA~NCEL"},
                   LBL_ALPHASEP = {'it': "Separatori alfabetici", 'fr': 'Séparateur alphabétique', 'en': "Alphabetical delimiter"},
                   LBL_ALL = {'it': "Applica a tutti", 'fr': 'Appliquer à tous', 'en': "Apply to all"},
                   LBL_STYLES = {'it': 'Stili', 'fr': 'Styles', 'en': "Styles"},
                   LBL_LEVELtitle = {'it': "Titolo", 'fr': "Titre", 'en': "Title"},
                   LBL_LEVEL0 = {'it': 'Separatore', 'fr': 'Séparateur', 'en': "Delimiter"},
                   LBL_LEVEL1 = {'it': 'Livello 1', 'fr': 'Niveau 1', 'en': "Level 1"},
                   LBL_LEVEL2 = {'it': 'Livello 2', 'fr': 'Niveau 2', 'en': "Level 2"},
                   LBL_LEVEL3 = {'it': 'Livello 3', 'fr': 'Niveau 3', 'en': "Level 3"},
                   LBL_DEFAULTSTYLES = {'it': 'Stili predefiniti', 'fr': 'Styles par défaut', 'en': "Default styles"},
                   MSG_INSERTED = {'it': 'indice "{}" inserito', 'fr': 'index {} inséré', 'en': "index {} inserted"},
                   MSG_CANCELLED = {'it': 'indice "{}" cancellato', 'fr': 'index {} supprimé', 'en': "index {} deleted"},
                   MSG_UPDATE = {'it': 'indice "{}" aggiornato', 'fr': 'index {} mis à jour', 'en': "index {} updated"})

    def __init__(self, ctx, doc, loc='en'):
        if not loc in self.availablelocales():
            loc = 'fr'
        self.loadlocale(loc)
        self.ctx = ctx
        self.smgr = ctx.ServiceManager
        self.doc = doc
        self.viewcursor = doc.CurrentController.ViewCursor
        try:
            self.currentpos = self.viewcursor.Start
            self.vctext = self.viewcursor.Text
        except UnknownPropertyException:
            return
        self.undo = self.doc.UndoManager
        self.text = doc.Text
        self.sections = doc.TextSections
        self.indexes = defaultdict(lambda: defaultdict(Entry))
        # load styles
        styles = doc.StyleFamilies.ParagraphStyles
        self.parastyles = {x: styles.getByName(x).DisplayName for x in styles.ElementNames}
        self.parastylesrev = {styles.getByName(x).DisplayName:x for x in styles.ElementNames}
        # execute dialog
        self.indexdialog()

    def indexdialog(self):
        try:
            # need this for inserting index into table
            table = self.viewcursor.TextTable
            if table:
                self.viewcursor.gotoRange(table.Anchor, False)
            self.doc.lockControllers()
            dialogprovider = self.smgr.createInstanceWithArguments("com.sun.star.awt.DialogProvider2", (self.doc,))
            uri = "$(user)/Scripts/python/multi_lex_index.dialog"
            dialog_url = _substitute_variables(uri, self.smgr)
            dialog = dialogprovider.createDialogWithHandler(dialog_url, self)
            dialog.setTitle(self.TITLE)
            # populate labels
            for name, label in zip(('lblCHOICE', 'lblTITLE', 'lblOPTIONS', 'btnOK', 'btnDELETE', 'cbALPHASEP', 'cbALL', 'btnCANCEL',
                                    'lblSTYLES', 'lblLEVEL0', 'lblLEVEL1', 'lblLEVEL2', 'lblLEVEL3', 'lblLEVELtitle', 'btnDEFAULTSTYLES'),
                                   (self.LBL_CHOICE, self.LBL_TITLE, self.LBL_OPTIONS, self.LBL_INSERT, self.LBL_DELETE, self.LBL_ALPHASEP, self.LBL_ALL, self.LBL_CANCEL,
                                    self.LBL_STYLES, self.LBL_LEVEL0, self.LBL_LEVEL1, self.LBL_LEVEL2, self.LBL_LEVEL3, self.LBL_LEVELtitle, self.LBL_DEFAULTSTYLES)):
                control = dialog.getControl(name)
                control.Model.Label = label
                if not name.startswith('lb'):
                    setattr(self, name, control)
            # populate styles
            itemlist = tuple(sorted(self.parastyles.values(), key=locale.strxfrm))
            for key in self.DEFAULTSTYLES.keys():
                lbox = dialog.getControl('lbLEVEL{}'.format(key))
                lbox.Model.StringItemList = itemlist
            self.getindexes(self.text)
            self.lbINDEXES = dialog.getControl("lbINDEXES")
            self.tfTITLE = dialog.getControl("tfTITLE")
            try:
                indexnames = sorted(self.indexes.keys(), key=locale.strxfrm)
            except:
                indexnames = sorted(self.indexes.keys())
            self.lbINDEXES.addItems(tuple(indexnames), 0)
            if indexnames:
                pos = 0
                try:
                    indexname = self.viewcursor.TextSection.Name.split("_")[0]
                    pos = self.lbINDEXES.Items.index(indexname)
                except:
                    pass
                self.lbINDEXES.selectItemPos(pos, True)
            dialog.execute()
        finally:
            self.doc.unlockControllers()

    def createindex(self, indexname, attributes, title):
        if not indexname:
            return
        try:
            self.undo.enterUndoContext(self.MSG_INSERTED.format(indexname))
            # header
            headersection = self.insertsection(indexname+self.HEADER_SUFFIX)
            headersection.IsVisible = (title != "")
            headersection.Anchor.setString(title)
            headersection.Anchor.ParaStyleName = attributes['styles']['title']
            # content
            contentsection = self.insertsection(indexname+self.CONTENT_SUFFIX, **attributes)
            self.insertlines(contentsection, indexname, **attributes)
        except:
            traceback.print_exc()
        finally:
            self.undo.leaveUndoContext()

    def insertsection(self, name, alphasep=None, styles=None):
        section = self.doc.createInstance("com.sun.star.text.TextSection")
        section.setName(name)
        self.vctext.insertTextContent(self.viewcursor, section, True)
        # custom attributes to save parameters
        if alphasep != None:
            uda = section.UserDefinedAttributes
            uda.insertByName("alphasep", AttributeData("", "CDATA", alphasep))
            uda.insertByName("styles", AttributeData("", "CDATA", str(styles)))
            section.UserDefinedAttributes = uda
        return section

    def insertlines(self, section, indexname, alphasep, styles):
        def strxfrmy(seq):
            return [locale.strxfrm(x) for x in seq]

        def buildlines(entries_, level=1):
            if level == 0:   # alphabetical delimiters
                key = lambda x: x[0][0]
            else:
                key = itemgetter(*range(level))
            for g, items in groupby(entries_, key):
                if level <= 1:
                    g = (g,)
                insertline(g, level)
                items = [i for i in items if len(i) > level]
                if items:
                    buildlines(items, level+1)

        def insertline(entry, level):
            section.Anchor.End.ParaStyleName = styles[level]
            if entry in index:  # entry with pages
                line = self.PATTERN.format(entry[-1])
                T.insertString(section.Anchor.End, line, False)
                if index[entry].ismain:
                    section.Anchor.End.CharStyleName = "Main index entry"
                pages = ', '.join(self.getpagerange(index[entry].pages))
                T.insertString(section.Anchor.End, pages, False)
            else:   # keys without pages
                line = entry[level-1]
                T.insertString(section.Anchor.End, line, False)
            T.insertControlCharacter(section.Anchor.End, 5, False)

        index = self.indexes[indexname]
        try:
            entries = sorted(index.keys(), key=strxfrmy)
        except:
            entries = sorted(index.keys())
        T = section.Anchor.Text
        if alphasep:
            buildlines(entries, 0)   # with alphabetical delimiters
        else:
            buildlines(entries, 1)

    def updateallindexes(self, indexnames):
        try:
            for indexname in indexnames:
                if self.sections.hasByName(indexname+self.CONTENT_SUFFIX):
                    self.updateindex(indexname)
        except:
            traceback.print_exc()

    def updateindex(self, indexname, attributes=None, title=None):
        if not indexname:
            return
        try:
            self.undo.enterUndoContext(self.MSG_UPDATE.format(indexname))
            # header
            if title != None:
                headersection = self.sections.getByName(indexname+self.HEADER_SUFFIX)
                headersection.IsVisible = (title != "")
                headersection.Anchor.setString(title)
                headersection.Anchor.ParaStyleName = attributes['styles']['title']
            # content
            if attributes == None:
                attributes = self.get_userattributes(indexname+self.CONTENT_SUFFIX)
            else:
                self.update_userattributes(indexname, **attributes)
            contentsection = self.sections.getByName(indexname+self.CONTENT_SUFFIX)
            self.viewcursor.gotoRange(contentsection.Anchor, False)
            contentsection.Anchor.setString('')
            self.insertlines(contentsection, indexname, **attributes)
        except:
            traceback.print_exc()
        finally:
            self.undo.leaveUndoContext()

    def deleteallindexes(self, indexnames):
        try:
            for indexname in indexnames:
                if self.sections.hasByName(indexname+self.CONTENT_SUFFIX):
                    self.deleteindex(indexname)
        except:
            traceback.print_exc()

    def deleteindex(self, indexname):
        try:
            self.undo.enterUndoContext(self.MSG_CANCELLED.format(indexname))
            for name in (indexname+self.HEADER_SUFFIX, indexname+self.CONTENT_SUFFIX,):
                try:
                    section = self.sections.getByName(name)
                    anchor = section.Anchor
                    self.viewcursor.gotoRange(anchor.Start, False)
                    anchor.setString("")
                    section.dispose() 
                    c = anchor.Text.createTextCursor()
                    if not c.TextTable:
                        anchor.createEnumeration().nextElement().dispose()
                except NoSuchElementException:
                    pass
        finally:
            self.undo.leaveUndoContext()

    def getindexes(self, text):
        viewcursor = self.viewcursor
        paraenum = text.createEnumeration()
        offset = 0
        while paraenum.hasMoreElements():
            para = paraenum.nextElement()
            if para.PageNumberOffset:
                self.viewcursor.gotoRange(para.Start, False)
                offset = para.PageNumberOffset - self.viewcursor.Page
            try:
                portionenum = para.createEnumeration()
                while portionenum.hasMoreElements():
                    portion = portionenum.nextElement()
                    try:
                        mark = portion.DocumentIndexMark
                        if mark:
                            index = self.indexes[mark.UserIndexName]
                            if mark.MarkEntry:
                                markentry = mark.MarkEntry.strip()
                                # entries = tuple(self.capitalize(x.strip(' '+self.ISMAIN)) for x in markentry.split(self.SEP, self.MAXKEYS))
                                entries = tuple(self.capitalize(x) for x in markentry.rstrip(self.ISMAIN).split(self.SEP, self.MAXKEYS))
                                viewcursor.gotoRange(portion, False)
                                index[entries].add(viewcursor.Page+offset)
                                if markentry.endswith(self.ISMAIN):
                                    index[entries].ismain = True
                            else:
                                if portion.IsStart:
                                    viewcursor.gotoRange(portion.Start, False)
                                else:
                                    viewcursor.gotoRange(portion.Start, True)
                                    index[(self.capitalize(viewcursor.String),)].add(viewcursor.Page+offset)
                    except AttributeError:
                        pass
            except AttributeError:
                pass
        # check sub texts
        if text == self.text:
            frames = self.doc.TextFrames
            for n in range(frames.Count):
                frame = frames.getByIndex(n)
                self.getindexes(frame)
            tables = self.doc.TextTables
            for n in range(tables.Count):
                table = tables.getByIndex(n)
                for cellname in table.CellNames:
                    cell = table.getCellByName(cellname)
                    self.getindexes(cell)
        viewcursor.gotoRange(self.currentpos, False)

    def mergepages(self, indexname):
        '''obsolete'''
        indexmarks = self.indexes[indexname]
        for entry in indexmarks:
            pages = ', '.join(self.getpagerange(indexmarks[entry]))
            indexmarks[entry] = "{} \t {}".format(entry[-1], pages)

    def getpagerange(self, pages):
        data = []
        pages = sorted(pages)
        for k, g in groupby(enumerate(pages), lambda x: x[0]-x[1]):
            g = list(g)
            d = "{}".format(g[0][1])
            if len(g) > 1:
                d += "-{}".format(g[-1][1])
            data.append(d)
        return data

    def callHandlerMethod(self, dialog, event, method):
        def getcustomstyles():
            customstyles = {}
            for key in self.DEFAULTSTYLES.keys():
                lbox = dialog.getControl('lbLEVEL{}'.format(key))
                customstyles[key] = self.parastylesrev[lbox.SelectedItem]
            return customstyles

        def select_styles(styles=None):
            if not styles:
                styles = self.DEFAULTSTYLES
            for key in self.DEFAULTSTYLES.keys():
                lbox = dialog.getControl('lbLEVEL{}'.format(key))
                lbox.selectItem(self.parastyles[styles[key]], True)

        def toggle_controls(enable):
            '''Controls not allowed for all-at-once actions'''
            for control in ('cbALPHASEP', 'lbLEVELtitle', 'lbLEVEL0', 'lbLEVEL1', 'lbLEVEL2', 'lbLEVEL3', 'btnDEFAULTSTYLES', 'tfTITLE'):
                dialog.getControl(control).setEnable(enable)

        try:
            if method == 'choice':
                choice = event.Source.SelectedItem
                content = choice + self.CONTENT_SUFFIX
                header = choice + self.HEADER_SUFFIX
                self.btnOK.setEnable(True)
                toggle_controls(True)
                exists = isinstance(self.sections.ElementNames, tuple) and content in self.sections.ElementNames
                self.cbALL.setEnable(exists)
                self.cbALL.State = 0
                if exists:
                    attributes = self.get_userattributes(content)
                    self.btnOK.setLabel(self.LBL_UPDATE)
                    self.btnDELETE.setEnable(True)
                    if 'alphasep' in attributes:
                        self.cbALPHASEP.State = attributes['alphasep']
                    self.tfTITLE.setText(self.sections.getByName(header).Anchor.String)
                    if 'styles' in attributes:
                        select_styles(attributes['styles'])
                    else:
                        select_styles(self.DEFAULTSTYLES)
                else:
                    self.btnOK.setLabel(self.LBL_INSERT)
                    self.btnDELETE.setEnable(False)
                    self.cbALPHASEP.State = 0
                    self.tfTITLE.setText(self.capitalize(choice))
                    select_styles()
            elif method == "toall":
                toggle_controls(not event.Selected)
            elif method == "OK":
                choice = self.lbINDEXES.SelectedItem
                title = self.tfTITLE.Text.strip()
                if self.sections.hasByName(choice+self.CONTENT_SUFFIX):
                    applytoall = self.cbALL.State
                    if applytoall:
                        self.updateallindexes(self.lbINDEXES.Items)
                    else:
                        attributes = dict(alphasep=self.cbALPHASEP.State, styles=getcustomstyles())
                        self.updateindex(choice, attributes, title)
                else:
                    attributes = dict(alphasep=self.cbALPHASEP.State, styles=getcustomstyles())
                    self.createindex(choice, attributes, title)
                dialog.endExecute()
            elif method == "delete":
                choice = self.lbINDEXES.SelectedItem
                if self.cbALL.State==1:
                    self.deleteallindexes(self.lbINDEXES.Items)
                else:
                    self.deleteindex(choice)
                dialog.endExecute()
            elif method == 'defaultstyles':
                select_styles()
            else:
                return False
        except:
            traceback.print_exc()

    def getSupportedMethodNames(self):
        return 'choice', 'toall', 'OK', 'delete', 'defaultstyles'

    def get_userattributes(self, sectionname):
        uda = self.sections.getByName(sectionname).UserDefinedAttributes
        attributes = {}
        for name in uda.ElementNames:
            attributes[name] = eval(uda.getByName(name).Value)
        return attributes

    def update_userattributes(self, sectionname, alphasep, styles):
        section = self.sections.getByName(sectionname+self.CONTENT_SUFFIX)
        uda = section.UserDefinedAttributes
        # check first if attribute exists, in case of corrupted section
        if uda.hasByName("alphasep"):
            if uda.getByName("alphasep").Value != str(alphasep):
                uda.replaceByName("alphasep", AttributeData("", "CDATA", str(alphasep)))
        else:
            uda.insertByName("alphasep", AttributeData("", "CDATA", str(alphasep)))
        if uda.hasByName("styles"):
            if uda.getByName("styles").Value != str(styles):
                uda.replaceByName("styles", AttributeData("", "CDATA", str(styles)))
        else:
            uda.insertByName("styles", AttributeData("", "CDATA", str(styles)))
        section.UserDefinedAttributes = uda

    def update_alphasep(self, sectionname, alphasep):
        section = self.sections.getByName(sectionname+self.CONTENT_SUFFIX)
        uda = section.UserDefinedAttributes
        if uda.getByName("alphasep").Value != str(alphasep):
            uda.replaceByName("alphasep", AttributeData("", "CDATA", str(alphasep)))
            section.UserDefinedAttributes = uda

    def update_styles(self, sectionname, styles):
        styles = str(styles)
        section = self.sections.getByName(sectionname+self.CONTENT_SUFFIX)
        uda = section.UserDefinedAttributes
        if uda.getByName("styles").Value != styles:
            uda.replaceByName("styles", AttributeData("", "CDATA", styles))
            section.UserDefinedAttributes = uda

    def capitalize(self, s):
        return s[0].upper() + s[1:]

    def loadlocale(self, loc):
        for k, v in self.LOCALES.items():
            setattr(self, k, v[loc])

    def availablelocales(self):
        return self.LOCALES['TITLE'].keys()


def insertindex(event=None):
    ctx = XSCRIPTCONTEXT.getComponentContext()
    doc = XSCRIPTCONTEXT.getDocument()
    _loc = _substitute_variables("$(vlang)", ctx.ServiceManager, False)
    loc = _loc.split('-')[0]
    if doc.supportsService("com.sun.star.text.GenericTextDocument"):
        indexcreator = IndexCreator(ctx, doc, loc)


def _substitute_variables(s, smgr, required=True):
    return smgr.createInstance("com.sun.star.util.PathSubstitution").substituteVariables(s, required)


g_exportedScripts = insertindex,
